/*
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#import "HDTVirtualMemory.h"

@implementation HDTVirtualMemory

static void InitializeByteOrder(int *byteOrder)
{
	if (*byteOrder == NATIVE_ENDIAN_BYTE_ORDER)
	{
		*byteOrder = (CFByteOrderGetCurrent() == CFByteOrderBigEndian) ? (BIG_ENDIAN_BYTE_ORDER) : (LITTLE_ENDIAN_BYTE_ORDER);
	}
}

+ (NSNumber *)readInt8AtProcess:(pid_t)process
				  memoryAddress:(mach_vm_address_t)address
{
	SInt8 signedInt8;
	mach_vm_size_t size = sizeof(SInt8);
	
	if (VMReadBytes(process, address, &signedInt8, &size))
	{
		return [NSNumber numberWithChar:signedInt8];
	}
  
	return nil;
}

+ (BOOL)writeInt8AtProcess:(pid_t)process
             memoryAddress:(mach_vm_address_t)address
                     value:(SInt8)value
{
	return VMWriteBytes(process, address, &value, sizeof(SInt8));
}

+ (NSNumber *)readUnsignedInt8AtProcess:(pid_t)process
						  memoryAddress:(mach_vm_address_t)address
{
	UInt8 unsignedInt8;
	mach_vm_size_t size = sizeof(UInt8);
	
	if (VMReadBytes(process, address, &unsignedInt8, &size))
	{
		return [NSNumber numberWithUnsignedChar:unsignedInt8];
	}
	
	return nil;
}

+ (BOOL)writeUnsignedInt8AtProcess:(pid_t)process
                     memoryAddress:(mach_vm_address_t)address
                             value:(UInt8)value
{
	return VMWriteBytes(process, address, &value, sizeof(UInt8));
}

+ (NSNumber *)readInt16AtProcess:(pid_t)process
				   memoryAddress:(mach_vm_address_t)address
					   byteOrder:(int)byteOrder
{
	SInt16 signedInt16;
	mach_vm_size_t size = sizeof(SInt16);
	
	if (VMReadBytes(process, address, &signedInt16, &size))
	{
		InitializeByteOrder(&byteOrder);
		
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			signedInt16 = CFSwapInt16BigToHost(signedInt16);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			signedInt16 = CFSwapInt16LittleToHost(signedInt16);
		}
		return [NSNumber numberWithShort:signedInt16];
	}
  
	return nil;
}

+ (BOOL)writeInt16AtProcess:(pid_t)process
              memoryAddress:(mach_vm_address_t)address
                      value:(SInt16)value
				  byteOrder:(int)byteOrder
{
	InitializeByteOrder(&byteOrder);
	
	if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt16HostToBig(value);
	}
	else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt16HostToLittle(value);
	}
	return VMWriteBytes(process, address, &value, sizeof(SInt16));
}

+ (NSNumber *)readUnsignedInt16AtProcess:(pid_t)process
				                   memoryAddress:(mach_vm_address_t)address
							   byteOrder:(int)byteOrder
{
	UInt16 unsignedInt16;
	mach_vm_size_t size = sizeof(UInt16);
	
	if (VMReadBytes(process, address, &unsignedInt16, &size))
	{
		InitializeByteOrder(&byteOrder);
		
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			unsignedInt16 = CFSwapInt16BigToHost(unsignedInt16);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			unsignedInt16 = CFSwapInt16LittleToHost(unsignedInt16);
		}
		return [NSNumber numberWithUnsignedShort:unsignedInt16];
	}
  
	return nil;
}

+ (BOOL)writeUnsignedInt16AtProcess:(pid_t)process
                      memoryAddress:(mach_vm_address_t)address
                              value:(UInt16)value
						  byteOrder:(int)byteOrder
{
	InitializeByteOrder(&byteOrder);
	
	if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt16HostToBig(value);
	}
	else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt16HostToLittle(value);
	}
	return VMWriteBytes(process, address, &value, sizeof(UInt16));
}

+ (NSNumber *)readInt32AtProcess:(pid_t)process
				   memoryAddress:(mach_vm_address_t)address
				       byteOrder:(int)byteOrder
{
	SInt32 signedInt32;
	mach_vm_size_t size = sizeof(SInt32);
	
	if (VMReadBytes(process, address, &signedInt32, &size))
	{
		InitializeByteOrder(&byteOrder);
		
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			signedInt32 = CFSwapInt32BigToHost(signedInt32);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			signedInt32 = CFSwapInt32LittleToHost(signedInt32);
		}
		return [NSNumber numberWithLong:signedInt32];
	}
  
	return nil;
}

+ (BOOL)writeInt32AtProcess:(pid_t)process
              memoryAddress:(mach_vm_address_t)address
                      value:(SInt32)value
				  byteOrder:(int)byteOrder
{
	InitializeByteOrder(&byteOrder);
	
	if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt32HostToBig(value);
	}
	else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt32HostToLittle(value);
	}
	return VMWriteBytes(process, address, &value, sizeof(SInt32));
}

+ (NSNumber *)readUnsignedInt32AtProcess:(pid_t)process
						   memoryAddress:(mach_vm_address_t)address
							   byteOrder:(int)byteOrder
{
	UInt32 unsignedInt32;
	mach_vm_size_t size = sizeof(UInt32);
	
	if (VMReadBytes(process, address, &unsignedInt32, &size))
	{
		InitializeByteOrder(&byteOrder);
		
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			unsignedInt32 = CFSwapInt32BigToHost(unsignedInt32);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			unsignedInt32 = CFSwapInt32LittleToHost(unsignedInt32);
		}
		return [NSNumber numberWithUnsignedLong:unsignedInt32];
	}
	
	return nil;
}

+ (BOOL)writeUnsignedInt32AtProcess:(pid_t)process
                      memoryAddress:(mach_vm_address_t)address
                              value:(UInt32)value
						  byteOrder:(int)byteOrder
{
	InitializeByteOrder(&byteOrder);
	
	if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt32HostToBig(value);
	}
	else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt32HostToLittle(value);
	}
	return VMWriteBytes(process, address, &value, sizeof(UInt32));
}

+ (NSNumber *)readInt64AtProcess:(pid_t)process
				   memoryAddress:(mach_vm_address_t)address
					   byteOrder:(int)byteOrder
{
	SInt64 signedInt64;
	mach_vm_size_t size = sizeof(SInt64);
	
	if (VMReadBytes(process, address, &signedInt64, &size))
	{
		InitializeByteOrder(&byteOrder);
		
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			signedInt64 = CFSwapInt64BigToHost(signedInt64);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			signedInt64 = CFSwapInt64LittleToHost(signedInt64);
		}
		return [NSNumber numberWithLongLong:signedInt64];
	}
	
	return nil;
}

+ (BOOL)writeInt64AtProcess:(pid_t)process
              memoryAddress:(mach_vm_address_t)address
                      value:(SInt64)value
				  byteOrder:(int)byteOrder
{
	InitializeByteOrder(&byteOrder);
	
	if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt64HostToBig(value);
	}
	else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt64HostToLittle(value);
	}
	return VMWriteBytes(process, address, &value, sizeof(SInt64));
}

+ (NSNumber *)readUnsignedInt64AtProcess:(pid_t)process
						   memoryAddress:(mach_vm_address_t)address
							   byteOrder:(int)byteOrder
{
	UInt64 unsignedInt64;
	mach_vm_size_t size = sizeof(UInt64);
	
	if (VMReadBytes(process, address, &unsignedInt64, &size))
	{
		InitializeByteOrder(&byteOrder);
		
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			unsignedInt64 = CFSwapInt64BigToHost(unsignedInt64);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			unsignedInt64 = CFSwapInt64LittleToHost(unsignedInt64);
		}
		return [NSNumber numberWithUnsignedLongLong:unsignedInt64];
	}
	
	return nil;
}

+ (BOOL)writeUnsignedInt64AtProcess:(pid_t)process
                      memoryAddress:(mach_vm_address_t)address
                              value:(UInt64)value
						  byteOrder:(int)byteOrder
{
	InitializeByteOrder(&byteOrder);
	
	if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt64HostToBig(value);
	}
	else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
	{
		value = CFSwapInt64HostToLittle(value);
	}
	return VMWriteBytes(process, address, &value, sizeof(UInt64));
}

+ (NSNumber *)readFloatAtProcess:(pid_t)process
                   memoryAddress:(mach_vm_address_t)address
					   byteOrder:(int)byteOrder
{
	float floatValue;
	mach_vm_size_t size = sizeof(float);
	
	if (VMReadBytes(process, address, &floatValue, &size))
	{
		InitializeByteOrder(&byteOrder);
		
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			int newValue = CFSwapInt32BigToHost(*((int *)&floatValue));
			floatValue = *((float *)&newValue);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			int newValue = CFSwapInt32LittleToHost(*((int *)&floatValue));
			floatValue = *((float *)&newValue);
		}
		return [NSNumber numberWithFloat:floatValue];
	}
	
	return nil;
}

+ (BOOL)writeFloatAtProcess:(pid_t)process
              memoryAddress:(mach_vm_address_t)address
                      value:(float)value
				  byteOrder:(int)byteOrder
{
	InitializeByteOrder(&byteOrder);
	
	if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
	{
		int newValue = CFSwapInt32HostToBig(*((int *)&value));
		value = *((float *)&newValue);
	}
	else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
	{
		int newValue = CFSwapInt32HostToLittle(*((int *)&value));
		value = *((float *)&newValue);
	}
	return VMWriteBytes(process, address, &value, sizeof(float));
}

+ (NSNumber *)readDoubleAtProcess:(pid_t)process
                    memoryAddress:(mach_vm_address_t)address
						byteOrder:(int)byteOrder
{
	double doubleValue;
	mach_vm_size_t size = sizeof(double);
	
	if (VMReadBytes(process, address, &doubleValue, &size))
	{
		InitializeByteOrder(&byteOrder);
		
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			SInt64 newValue = CFSwapInt64BigToHost(*((SInt64 *)&doubleValue));
			doubleValue = *((double *)&newValue);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			SInt64 newValue = CFSwapInt64LittleToHost(*((SInt64 *)&doubleValue));
			doubleValue = *((double *)&newValue);
		}
		return [NSNumber numberWithDouble:doubleValue];
	}
	
	return nil;
}

+ (BOOL)writeDoubleAtProcess:(pid_t)process
               memoryAddress:(mach_vm_address_t)address
                       value:(double)value
				   byteOrder:(int)byteOrder
{
	InitializeByteOrder(&byteOrder);
	
	if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
	{
		SInt64 newValue = CFSwapInt64HostToBig(*((SInt64 *)&value));
		value = *((double *)&newValue);
	}
	else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
	{
		SInt64 newValue = CFSwapInt64HostToLittle(*((SInt64 *)&value));
		value = *((double *)&newValue);
	}
	return VMWriteBytes(process, address, &value, sizeof(double));
}

+ (NSString *)readUTF8StringAtProcess:(pid_t)process
						memoryAddress:(mach_vm_address_t)address
{
	NSMutableData *mutableData = [[NSMutableData alloc] init];
	while (YES)
	{
		mach_vm_size_t size = 1;
		char charValue = 0;
		if (!VMReadBytes(process, address, &charValue, &size))
		{
			[mutableData release];
			return nil;
		}
		
		[mutableData appendBytes:&charValue
						  length:1];
		
		if (charValue == '\0')
		{
			break;
		}
		
		address++;
	}
	
	NSString *stringRepresentation = [[NSString alloc] initWithData:mutableData
														   encoding:NSUTF8StringEncoding];
	[mutableData release];
	return [stringRepresentation autorelease];
}

+ (BOOL)writeUTF8StringAtProcess:(pid_t)process
				   memoryAddress:(mach_vm_address_t)address
						   value:(NSString *)value
{
	const char *cString = [value cStringUsingEncoding:NSUTF8StringEncoding];
	return VMWriteBytes(process, address, cString, strlen(cString) + 1);
}

+ (NSString *)readUTF16StringAtProcess:(pid_t)process
						 memoryAddress:(mach_vm_address_t)address
							 byteOrder:(int)byteOrder
{
	NSMutableData *mutableData = [[NSMutableData alloc] init];
	
	InitializeByteOrder(&byteOrder);
	
	while (YES)
	{
		short shortValue = 0;
		mach_vm_size_t size = sizeof(short);
		
		if (!VMReadBytes(process, address, &shortValue, &size))
		{
			[mutableData release];
			return nil;
		}
		
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			shortValue = CFSwapInt16BigToHost(shortValue);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			shortValue = CFSwapInt16LittleToHost(shortValue);
		}
		 
		[mutableData appendBytes:&shortValue
						  length:sizeof(short)];
		
		if (shortValue == 0)
		{
			break;
		}
		
		address += sizeof(short);
	}
	
	NSString *stringRepresentation = [NSString stringWithCharacters:[mutableData bytes]
															 length:[mutableData length] / 2];
	
	[mutableData release];
	return [stringRepresentation autorelease];
}

+ (BOOL)writeUTF16StringAtProcess:(pid_t)process
                   memoryAddress:(mach_vm_address_t)address
							value:(NSString *)value
						byteOrder:(int)byteOrder
{
	int size = [value length];
	unichar buffer[size + 1];
	[value getCharacters:buffer];
	buffer[size] = 0;
	
	InitializeByteOrder(&byteOrder);
	
	int bufferIndex;
	for (bufferIndex = 0; bufferIndex < size; bufferIndex++)
	{
		if (byteOrder == BIG_ENDIAN_BYTE_ORDER)
		{
			buffer[bufferIndex] = CFSwapInt16HostToBig(buffer[bufferIndex]);
		}
		else if (byteOrder == LITTLE_ENDIAN_BYTE_ORDER)
		{
			buffer[bufferIndex] = CFSwapInt16HostToLittle(buffer[bufferIndex]);
		}
	}
	
	return VMWriteBytes(process, address, buffer, sizeof(unichar) * size + sizeof(unichar));
}

@end
